use std::{error::Error, str::FromStr, time::Duration};

use crate::{handling::response::CookieBuilder, utils::urlencoded::DeCodable};

/// Structure representing a Cookie
/// # Creating a Cookie:
/// ```
/// use http::handling::response::Cookie;
/// use http::handling::response::CookieBuilder;
///
/// let cookie = CookieBuilder::build("name", "value").finish();
/// ```
#[derive(PartialEq, Eq, PartialOrd, Ord, Debug)]
pub struct Cookie {
    /// Storage for the cookie string. Only used if this structure was derived
    /// from a string that was subsequently parsed.
    pub(crate) cookie_string: Option<String>,
    pub(crate) name: String,
    pub(crate) value: String,
    // expires: Option<Tm>,
    pub(crate) max_age: Option<Duration>,
    /// The cookie's domain, if any.
    pub(crate) domain: Option<String>,
    /// The cookie's path domain, if any.
    pub(crate) path: Option<String>,
    /// Whether this cookie was marked Secure.
    pub(crate) secure: bool,
    /// Whether this cookie was marked HttpOnly.
    pub(crate) http_only: bool,
    /// The draft `SameSite` attribute.
    pub(crate) same_site: Option<SameSite>,
    pub(crate) expires: Option<String>,
    pub(crate) partitioned: bool,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
/// SameSite Paremeters
pub enum SameSite {
    /// Requires Secure
    None,
    Lax,
    Strict,
}

impl std::fmt::Display for SameSite {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::None => write!(f, "SameSite=None"),
            Self::Lax => write!(f, "SameSite=Lax"),
            Self::Strict => write!(f, "SameSite=Strict"),
        }
    }
}

impl std::fmt::Display for Cookie {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut appendix = String::from("");
        if self.secure {
            appendix += "; Secure";
        }
        if self.http_only {
            appendix += "; HttpOnly";
        }
        if self.partitioned {
            appendix += "; Partitioned";
        }
        if let Some(max_age) = &self.max_age {
            appendix += &format!("; Max-Age={}", max_age.as_secs());
        }
        if let Some(domain) = &self.domain {
            appendix += &format!("; Domain={}", domain);
        }
        if let Some(path) = &self.path {
            appendix += &format!("; Path={}", path);
        }
        if let Some(same_site) = &self.same_site {
            appendix += &format!("; {}", same_site);
            if !self.secure && *same_site == SameSite::None {
                appendix += &format!("; Secure");
            }
        }
        if let Some(expires) = &self.expires {
            appendix += &format!("; Expires={}", expires)
        }
        write!(f, "Set-Cookie: {}={}{}", self.name, self.value, appendix)
    }
}

impl Error for ParseCookieError {}

#[derive(Debug)]
pub struct ParseCookieError {
    inner: CookieError,
}

#[derive(Debug, PartialEq, PartialOrd, Eq, Ord)]
pub enum CookieError {
    MissingEqual,
    InvalidAttribute,
}

impl std::fmt::Display for ParseCookieError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "ParseCookieError {{ error: {:?} }}", self.inner)
    }
}

impl FromStr for Cookie {
    type Err = ParseCookieError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut final_result = CookieBuilder::build("", "");
        let mut first = true;
        for part in s.split(';') {
            let trimmed_part = part.trim();
            if first {
                let Some(name_val) = part.split_once('=') else {
                    return Err(Self::Err { inner: CookieError::MissingEqual  });
                };
                unsafe {
                    final_result = CookieBuilder::build(
                        &String::from_utf8_unchecked(if let Ok(name) = name_val.0.decode() {
                            name
                        } else {
                            name_val.0.into()
                        }),
                        &String::from_utf8_unchecked(if let Ok(value) = name_val.1.decode() {
                            value
                        } else {
                            name_val.1.into()
                        }),
                    );
                }
                first = false;
                break;
            }
            final_result = match trimmed_part {
                "Secure" => final_result.secure(true),
                "HttpOnly" => final_result.http_only(true),
                "Partitioned" => final_result.partitioned(true),
                _ => {
                    return Err(Self::Err {
                        inner: CookieError::InvalidAttribute,
                    });
                }
            }
        }
        println!("{:?}", final_result);
        Ok(final_result.finish())
    }
}

#[cfg(test)]
mod test {
    use std::time::Duration;

    use crate::handling::response::CookieBuilder;

    use super::SameSite;

    #[test]
    fn test_cookie_to_string() {
        let test_cookie1 = CookieBuilder::build("a", "cookie").finish().to_string();
        let test_cookie1_res = "Set-Cookie: a=cookie";
        let test_cookie2 = CookieBuilder::build("a", "secure_cookie")
            .secure(true)
            .finish()
            .to_string();
        let test_cookie2_res = "Set-Cookie: a=secure_cookie; Secure";
        let test_cookie3 = CookieBuilder::build("ab", "ss")
            .max_age(Duration::from_secs(24))
            .domain("codecraft.com")
            .path("/")
            .same_site(SameSite::None)
            .http_only(true)
            .partitioned(true)
            .expires("Monday")
            .finish();
        let test_cookie3_res = "Set-Cookie: ab=ss; HttpOnly; Partitioned; \
            Max-Age=24; Domain=codecraft.com; Path=/; SameSite=None; Secure; Expires=Monday";

        assert_eq!(test_cookie1_res, test_cookie1);
        assert_eq!(test_cookie2_res, test_cookie2);
        assert_eq!(test_cookie3_res, test_cookie3.to_string());
    }
    #[test]
    fn cooki_from_string() {}
}
